//
//  Little cms
//  Copyright (C) 1998-1999 Marti Maria
//
// THIS SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL MARTI MARIA BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//     Input/Output


#include "lcms.h"
#include "icc34.h"
#include <stdlib.h>


cmsHPROFILE   LCMSEXPORT cmsOpenProfileFromFile(const char *ICCProfile, const char *sAccess);
cmsHPROFILE   LCMSEXPORT cmsOpenProfileFromMem(LPVOID MemPtr, DWORD dwSize);
cmsHPROFILE   LCMSEXPORT cmsCreateRGBProfile(LPcmsCIExyY WhitePoint,
                                        LPcmsCIExyYTRIPLE Primaries,
                                        LPGAMMATABLE TransferFunction[3]);

BOOL          LCMSEXPORT cmsCloseProfile(cmsHPROFILE hProfile);
LPLUT         cdecl      cmsReadICCLut(cmsHPROFILE hProfile, icTagSignature sig);
LPGAMMATABLE  cdecl      cmsReadICCGamma(cmsHPROFILE hProfile, icTagSignature sig);
BOOL          LCMSEXPORT cmsTakeMediaWhitePoint(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile);
BOOL          LCMSEXPORT cmsTakeMediaBlackPoint(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile);
BOOL          LCMSEXPORT cmsTakeIluminant(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile);
BOOL          LCMSEXPORT cmsTakeColorants(LPcmsCIEXYZTRIPLE Dest, cmsHPROFILE hProfile);
BOOL          LCMSEXPORT cmsIsTag(cmsHPROFILE hProfile, icTagSignature sig);


const char*   LCMSEXPORT cmsTakeProductName(cmsHPROFILE hProfile);

icColorSpaceSignature   LCMSEXPORT cmsGetPCS(cmsHPROFILE hProfile);
icColorSpaceSignature   LCMSEXPORT cmsGetColorSpace(cmsHPROFILE hProfile);
icProfileClassSignature LCMSEXPORT cmsGetDeviceClass(cmsHPROFILE hProfile);

int                     cdecl cmsIsLinear(WORD Table[], int nEntries);


// ------------------- implementation -----------------------------------


// This is the internal struct holding profile details.

typedef struct {

               FILE         *stream;       // Associated stream. If NULL,
                                           // tags are supposed to be in
                                           // memory rather than in a file.

               // Only most important items found in ICC profile

               icProfileClassSignature DeviceClass;
               icColorSpaceSignature   ColorSpace;
               icColorSpaceSignature   PCS;
               icRenderingIntent       RenderingIntent;
               icUInt32Number          flags;
               cmsCIEXYZ               Illuminant;

               // Dictionary

               int             TagCount;
               icTagSignature *TagNames;
               size_t         *TagSizes;

               // union {

               size_t         *TagOffsets;       // Only one makes sense, depending
               LPVOID         *TagPtrs;          // on 'stream' member.

               // };                             // pity my compiler doesn't allow
                                                 // unnamed unions

               char   PhysicalFile[MAX_PATH];    // For take info

              } ICCPROFILE, FAR* LPICCPROFILE;



//
//      Little-Endian to Big-Endian
//

#ifdef USE_BIG_ENDIAN
#define AdjustEndianess16(a)
#define AdjustEndianess32(a)
#define AdjustEndianessArray16(a, b)
#else

static
void AdjustEndianess16(LPBYTE pByte)
{
       BYTE tmp;

       tmp = pByte[0];
       pByte[0] = pByte[1];
       pByte[1] = tmp;
}

static
void AdjustEndianess32(LPBYTE pByte)
{
        BYTE temp1;
        BYTE temp2;

        temp1 = *pByte++;
        temp2 = *pByte++;
        *(pByte-1) = *pByte;
        *pByte++ = temp2;
        *(pByte-3) = *pByte;
        *pByte = temp1;
}

// swap bytes in a array of words

static
void AdjustEndianessArray16(LPWORD p, size_t num_words)
{
       swab((char*) p, (char*)p, (int) num_words * sizeof(WORD));
}

#endif


// Fixed point conversion

static
double Convert8Fixed8(WORD fixed8)
{
       BYTE msb, lsb;

       msb = (BYTE) (fixed8 & 0xff);
       lsb = (BYTE) (((WORD) fixed8 >> 8) & 0xff);

       return (double) ((double) msb + ((double) lsb / 256.0));
}

static
double Convert15Fixed16(icS15Fixed16Number fix32)
{
    double floater, sign;
    int Whole, Frac;


    AdjustEndianess32((LPBYTE) &fix32);

    sign  = (fix32 < 0 ? -1 : 1);
    fix32 = abs(fix32);

    Whole = LOWORD(fix32 >> 16);
    Frac  = LOWORD(fix32 & 0x0000ffffL);

    floater = (double) Whole + ((double) Frac / 65536.0);
    return sign*floater;
}



static
ICCPROFILE *ICCFileOpen(FILE *ICCfile)
{
       ICCPROFILE *Icc;
       icTag Tag;
       icHeader Header;
       long TagCount, i;

       fread(&Header, sizeof(icHeader), 1, ICCfile);

       // Convert endian

       AdjustEndianess32((LPBYTE) &Header.size);
       AdjustEndianess32((LPBYTE) &Header.cmmId);
       AdjustEndianess32((LPBYTE) &Header.version);
       AdjustEndianess32((LPBYTE) &Header.deviceClass);
       AdjustEndianess32((LPBYTE) &Header.colorSpace);
       AdjustEndianess32((LPBYTE) &Header.pcs);
       AdjustEndianess32((LPBYTE) &Header.magic);
       AdjustEndianess32((LPBYTE) &Header.flags);
       AdjustEndianess32((LPBYTE) &Header.renderingIntent);

       // Validate it

       if (Header.magic != icMagicNumber)
       {
              fclose(ICCfile);
              return NULL;
       }


       Icc = malloc(sizeof(ICCPROFILE));
       if (!Icc)                          // can't allocate
       {
       fclose(ICCfile);
       return NULL;
       }

       ZeroMemory(Icc, sizeof(ICCPROFILE));

       Icc -> stream = ICCfile;

       if (fread(&TagCount, sizeof(long), 1, ICCfile) != 1)
                     goto ErrorCleanup;

       AdjustEndianess32((LPBYTE) &TagCount);

       Icc -> DeviceClass     = Header.deviceClass;
       Icc -> ColorSpace      = Header.colorSpace;
       Icc -> PCS             = Header.pcs;
       Icc -> RenderingIntent = Header.renderingIntent;
       Icc -> Illuminant.X    = Convert15Fixed16(Header.illuminant.X);
       Icc -> Illuminant.Y    = Convert15Fixed16(Header.illuminant.Y);
       Icc -> Illuminant.Z    = Convert15Fixed16(Header.illuminant.Z);

       // Memory for tag tables

       Icc -> TagNames   = (icTagSignature *) malloc(sizeof(icTagSignature) * TagCount);
       Icc -> TagSizes   = (size_t *) malloc(sizeof(size_t) * TagCount);
       Icc -> TagOffsets = (size_t *) malloc(sizeof(size_t) * TagCount);
       Icc -> TagCount   = TagCount;

       for (i=0; i < TagCount; i++)
       {
              fread(&Tag, sizeof(icTag), 1, ICCfile);

              AdjustEndianess32((LPBYTE) &Tag.offset);
              AdjustEndianess32((LPBYTE) &Tag.size);
              AdjustEndianess32((LPBYTE) &Tag.sig);            // Signature

              Icc -> TagNames[i]   = Tag.sig;
              Icc -> TagOffsets[i] = Tag.offset;
              Icc -> TagSizes[i]   = Tag.size;

       }


       return Icc;


ErrorCleanup:

              fclose(ICCfile);
              free(Icc);
              return NULL;
}


static
void InitTag(int i, LPICCPROFILE Icc, icTagSignature sig, size_t size, LPVOID Init)
{
       LPVOID Ptr;

       Ptr = malloc(size);
       CopyMemory(Ptr, Init, size);

       Icc ->TagNames[i] = sig;
       Icc ->TagSizes[i] = size;
       Icc ->TagPtrs[i]  = Ptr;
}

// Duplicate a gamma table

static
LPGAMMATABLE DupGamma(LPGAMMATABLE In)
{
       LPGAMMATABLE Ptr;

       Ptr = cmsAllocGamma(In -> nEntries);
       CopyMemory(Ptr -> GammaTable,
                  In -> GammaTable,
                  In -> nEntries * sizeof(WORD));

       return Ptr;
}



static
int SizeOfGammaTab(LPGAMMATABLE In)
{
       return sizeof(GAMMATABLE) + (In -> nEntries - 1)*sizeof(WORD);
}


// This function creates a profile based on White point, primaries and
// transfer functions.

cmsHPROFILE LCMSEXPORT cmsCreateRGBProfile(LPcmsCIExyY WhitePoint,
                                LPcmsCIExyYTRIPLE Primaries,
                                LPGAMMATABLE TransferFunction[3])
{
       ICCPROFILE *Icc;
       cmsCIEXYZ tmp;
       MAT3 MColorants;
       cmsCIEXYZTRIPLE Colorants;

       Icc = malloc(sizeof(ICCPROFILE));
       if (!Icc)                          // can't allocate
       {
       return NULL;
       }

       ZeroMemory(Icc, sizeof(ICCPROFILE));

       Icc -> stream = NULL;
       Icc -> DeviceClass = icSigInputClass;
       Icc -> ColorSpace = icSigRgbData;
       Icc -> PCS = icSigXYZData;
       Icc -> RenderingIntent = icRelativeColorimetric;
       Icc -> flags = 0;
       Icc -> Illuminant.X    = 0.964294;
       Icc -> Illuminant.Y    = 1.0;
       Icc -> Illuminant.Z    = 0.825104;




       // I implement this profile as following tags:
       //
       //  1 icSigProfileDescriptionTag
       //  2 icSigMediaWhitePointTag
       //  3 icSigRedColorantTag
       //  4 icSigGreenColorantTag
       //  5 icSigBlueColorantTag
       //  6 icSigRedTRCTag
       //  7 icSigGreenTRCTag
       //  8 icSigBlueTRCTag
       //  9 icSigGrayTRCTag

       // This conforms a standard RGB InputProfile as says ICC


       Icc -> TagCount   = 8;
       // Memory for tag tables

       Icc -> TagNames   = (icTagSignature *) malloc(sizeof(icTagSignature) * Icc ->TagCount);
       Icc -> TagSizes   = (size_t *) malloc(sizeof(size_t) * Icc -> TagCount);
       Icc -> TagPtrs    = (LPVOID *) malloc(sizeof(LPVOID) * Icc -> TagCount);

       // Fill-in the tags

       InitTag(0, Icc, icSigProfileDescriptionTag, 11, "(internal)");
       cmsxyY2XYZ(&tmp, WhitePoint);

       InitTag(1, Icc, icSigMediaWhitePointTag, sizeof(cmsCIEXYZ), &tmp);

       if (!cmsBuildRGB2XYZtransferMatrix(&MColorants, WhitePoint, Primaries))
       {
              free(Icc -> TagPtrs);
              free(Icc -> TagSizes);
              free(Icc -> TagNames);
              free(Icc);
              return FALSE;
       }


       Colorants.Red.X = MColorants.v[0].n[0];
       Colorants.Red.Y = MColorants.v[1].n[0];
       Colorants.Red.Z = MColorants.v[2].n[0];

       Colorants.Green.X = MColorants.v[0].n[1];
       Colorants.Green.Y = MColorants.v[1].n[1];
       Colorants.Green.Z = MColorants.v[2].n[1];

       Colorants.Blue.X = MColorants.v[0].n[2];
       Colorants.Blue.Y = MColorants.v[1].n[2];
       Colorants.Blue.Z = MColorants.v[2].n[2];

       InitTag(2, Icc, icSigRedColorantTag,     sizeof(cmsCIEXYZ), &Colorants.Red);
       InitTag(3, Icc, icSigBlueColorantTag,    sizeof(cmsCIEXYZ), &Colorants.Blue);
       InitTag(4, Icc, icSigGreenColorantTag,   sizeof(cmsCIEXYZ), &Colorants.Green);

       // In case of gamma, we must dup' the table pointer

       InitTag(5, Icc, icSigRedTRCTag, SizeOfGammaTab(TransferFunction[0]), TransferFunction[0]);
       InitTag(6, Icc, icSigGreenTRCTag, SizeOfGammaTab(TransferFunction[1]), TransferFunction[1]);
       InitTag(7, Icc, icSigBlueTRCTag, SizeOfGammaTab(TransferFunction[2]), TransferFunction[2]);

       return (cmsHPROFILE) Icc;
}



// Create profile from ICC file
// TODO: For now, sAccess is ignored. Perhaps in a (far) future, I will consider
// the creation of profiles.

#ifdef __BORLANDC__
#pragma argsused
#endif

cmsHPROFILE LCMSEXPORT cmsOpenProfileFromFile(const char *lpFileName, const char *sAccess)
{
       LPICCPROFILE NewIcc;
       FILE *ICCfile;


       ICCfile = fopen(lpFileName, "rb");
       if (!ICCfile)
       {
              cmsSignalError(LCMS_ERRC_ABORTED, "File '%s' not found", lpFileName);
              return NULL;  // Urecoverable
       }

       NewIcc = ICCFileOpen(ICCfile);
       if (!NewIcc)
       {
              cmsSignalError(LCMS_ERRC_ABORTED, "Bad file format");
              return NULL;  // Urecoverable
       }

       strncpy(NewIcc -> PhysicalFile, lpFileName, MAX_PATH-1);

       // in 32 bit, simply typecast
       return (cmsHPROFILE) (LPSTR) NewIcc;
}

// Open from mem

cmsHPROFILE LCMSEXPORT cmsOpenProfileFromMem(LPVOID MemPtr, DWORD dwSize)
{
       LPICCPROFILE NewIcc;
       FILE *TempFile;


       TempFile = tmpfile();
       if (TempFile == NULL)
       {
              cmsSignalError(LCMS_ERRC_ABORTED, "Unable to create tempfile");
              return NULL;  // Urecoverable
       }

       if (fwrite(MemPtr, 1, dwSize, TempFile) !=
              dwSize)
       {
              cmsSignalError(LCMS_ERRC_ABORTED, "fwrite to tempfile failed");
              fclose(TempFile);
              return NULL;  // Urecoverable

       };

       rewind(TempFile);
       NewIcc = ICCFileOpen(TempFile);
       if (!NewIcc)
       {
                     cmsSignalError(LCMS_ERRC_ABORTED, "Error in embbedded profile");
                     fclose(TempFile);
                     return NULL;
       }

       return (cmsHPROFILE) (LPSTR) NewIcc;
}




BOOL LCMSEXPORT cmsCloseProfile(cmsHPROFILE hProfile)
{
       LPICCPROFILE icco = (LPICCPROFILE) (LPSTR) hProfile;
       FILE *file;

       if (!icco) return FALSE;

       file = icco -> stream;

       if (!file)
       {
              int i;

              for (i=0; i < icco -> TagCount; i++)
              {
                  if (icco -> TagPtrs[i])
                            free(icco -> TagPtrs[i]);
              }

              free(icco -> TagPtrs);

       }
       else   {
              fclose(file);
              free(icco -> TagOffsets);
              }

       free(icco -> TagNames);
       free(icco -> TagSizes);
       free(icco);

       return TRUE;
}


// Auxiliary

static
int SearchTag(LPICCPROFILE Profile, icTagSignature sig)
{
       int i;

       for (i=0; i < Profile -> TagCount; i++)
       {
              if (sig == Profile -> TagNames[i])
                            return i;
       }

       return -1;
}

// Is a table linear?


int cmsIsLinear(WORD Table[], int nEntries)
{
       register int i;

       for (i=0; i < nEntries; i++) {

       if ((double) Table[i] !=
              floor((((double) i / ((double) nEntries-1)) * 65535.0) + .5))
                     return 0;

       }

       return 1;
}


static
unsigned int uipow(unsigned int a, unsigned int b) {
        unsigned int rv = 1;
        for (; b > 0; b--)
                rv *= a;
        return rv;
}


// Convert between notations, as stated in ICC spec.

#define TO16(x)      (WORD) ((x) << 8)


// Case LUT 8

static
void ReadLUT8(LPICCPROFILE Icc, LPLUT NewLUT)
{
    static icLut8 LUT8;
    LPBYTE Temp;
    size_t nTabSize;
    unsigned int i, j;
    unsigned int AllLinear;
    LPWORD PtrW;

       fread(&LUT8, sizeof(icLut8) - SIZEOF_UINT8_ALIGNED, 1, Icc -> stream);

       // fread(&LUT8, sizeof(my_icLut8), 1, Icc -> stream);

       NewLUT -> wFlags        = LUT_HASTL1|LUT_HASTL2|LUT_HAS3DGRID;
       NewLUT -> cLutPoints    = LUT8.clutPoints;
       NewLUT -> InputChan     = LUT8.inputChan;
       NewLUT -> OutputChan    = LUT8.outputChan;
       NewLUT -> InputEntries  = 256;
       NewLUT -> OutputEntries = 256;


       AdjustEndianess32((LPBYTE) &LUT8.e00);
       AdjustEndianess32((LPBYTE) &LUT8.e01);
       AdjustEndianess32((LPBYTE) &LUT8.e02);
       AdjustEndianess32((LPBYTE) &LUT8.e10);
       AdjustEndianess32((LPBYTE) &LUT8.e11);
       AdjustEndianess32((LPBYTE) &LUT8.e12);
       AdjustEndianess32((LPBYTE) &LUT8.e20);
       AdjustEndianess32((LPBYTE) &LUT8.e21);
       AdjustEndianess32((LPBYTE) &LUT8.e22);


       // Matrix handling

       NewLUT -> Matrix.v[0].n[0] = (Fixed32) LUT8.e00;
       NewLUT -> Matrix.v[0].n[1] = (Fixed32) LUT8.e01;
       NewLUT -> Matrix.v[0].n[2] = (Fixed32) LUT8.e02;
       NewLUT -> Matrix.v[1].n[0] = (Fixed32) LUT8.e10;
       NewLUT -> Matrix.v[1].n[1] = (Fixed32) LUT8.e11;
       NewLUT -> Matrix.v[1].n[2] = (Fixed32) LUT8.e12;
       NewLUT -> Matrix.v[2].n[0] = (Fixed32) LUT8.e20;
       NewLUT -> Matrix.v[2].n[1] = (Fixed32) LUT8.e21;
       NewLUT -> Matrix.v[2].n[2] = (Fixed32) LUT8.e22;


       // Only operates if not identity...

       if (!MAT3isIdentity(&NewLUT -> Matrix, 0.0001)) {

              NewLUT -> wFlags |= LUT_HASMATRIX;
       }


       // Copy input tables

       Temp = malloc(256);
       AllLinear = 0;
       for (i=0; i < NewLUT -> InputChan; i++) {

              PtrW = (LPWORD) malloc(sizeof(WORD) * 256);
              NewLUT -> L1[i] = PtrW;
              fread(Temp, 1, 256, Icc -> stream);
              for (j=0; j < 256; j++)
                     PtrW[j] = TO16(Temp[j]);
                     AllLinear += cmsIsLinear(NewLUT -> L1[i], NewLUT -> InputEntries);
              }

       // Linear input, so ignore full step

       if (AllLinear == NewLUT -> InputChan) {

              NewLUT -> wFlags &= ~LUT_HASTL1;
       }

       free(Temp);

       // Copy 3D CLUT

       nTabSize = (NewLUT -> OutputChan * uipow(NewLUT->cLutPoints,
                                                NewLUT->InputChan));

       PtrW = (LPWORD) malloc(sizeof(WORD) * nTabSize);
       Temp = malloc(nTabSize);
       fread(Temp, 1, nTabSize, Icc -> stream);

       NewLUT -> T = PtrW;
       for (i = 0; i < nTabSize; i++) {

                     *PtrW++ = TO16(Temp[i]);
       }
       free(Temp);


       // Copy output tables

       Temp = malloc(256);
       AllLinear = 0;
       for (i=0; i < NewLUT -> OutputChan; i++) {

              PtrW = malloc(sizeof(WORD) * 256);
              NewLUT -> L2[i] = PtrW;
              fread(Temp, 1, 256, Icc -> stream);
              for (j=0; j < 256; j++)
                     PtrW[j] = TO16(Temp[j]);
                     AllLinear += cmsIsLinear(NewLUT -> L2[i], 256);
              }

       // Linear input, so ignore full step

       if (AllLinear == NewLUT -> InputChan) {

              NewLUT -> wFlags &= ~LUT_HASTL2;
       }


       free(Temp);

       cmsCalcL16Params(NewLUT -> InputEntries,  &NewLUT -> In16params);
       cmsCalcL16Params(NewLUT -> OutputEntries, &NewLUT -> Out16params);
       cmsCalcCLUT16Params(NewLUT -> cLutPoints,  NewLUT -> InputChan,
                                                  NewLUT -> OutputChan,
                                                  &NewLUT -> CLut16params);

}





// Case LUT 16

static
void ReadLUT16(LPICCPROFILE Icc, LPLUT NewLUT)
{
    static icLut16 LUT16;
    size_t nTabSize;
    unsigned int i;
    unsigned int AllLinear;
    LPWORD PtrW;


       fread(&LUT16, sizeof(icLut16)- SIZEOF_UINT16_ALIGNED, 1, Icc -> stream);

       // fread(&LUT16, sizeof(my_icLut16), 1, Icc -> stream);

       NewLUT -> wFlags        = LUT_HASTL1 | LUT_HASTL2 | LUT_HAS3DGRID;
       NewLUT -> cLutPoints    = LUT16.clutPoints;
       NewLUT -> InputChan     = LUT16.inputChan;
       NewLUT -> OutputChan    = LUT16.outputChan;

       AdjustEndianess16((LPBYTE) &LUT16.inputEnt);
       AdjustEndianess16((LPBYTE) &LUT16.outputEnt);

       NewLUT -> InputEntries  = LUT16.inputEnt;
       NewLUT -> OutputEntries = LUT16.outputEnt;


       // Matrix handling

       AdjustEndianess32((LPBYTE) &LUT16.e00);
       AdjustEndianess32((LPBYTE) &LUT16.e01);
       AdjustEndianess32((LPBYTE) &LUT16.e02);
       AdjustEndianess32((LPBYTE) &LUT16.e10);
       AdjustEndianess32((LPBYTE) &LUT16.e11);
       AdjustEndianess32((LPBYTE) &LUT16.e12);
       AdjustEndianess32((LPBYTE) &LUT16.e20);
       AdjustEndianess32((LPBYTE) &LUT16.e21);
       AdjustEndianess32((LPBYTE) &LUT16.e22);

       NewLUT -> Matrix.v[0].n[0] = (Fixed32) LUT16.e00;
       NewLUT -> Matrix.v[0].n[1] = (Fixed32) LUT16.e01;
       NewLUT -> Matrix.v[0].n[2] = (Fixed32) LUT16.e02;
       NewLUT -> Matrix.v[1].n[0] = (Fixed32) LUT16.e10;
       NewLUT -> Matrix.v[1].n[1] = (Fixed32) LUT16.e11;
       NewLUT -> Matrix.v[1].n[2] = (Fixed32) LUT16.e12;
       NewLUT -> Matrix.v[2].n[0] = (Fixed32) LUT16.e20;
       NewLUT -> Matrix.v[2].n[1] = (Fixed32) LUT16.e21;
       NewLUT -> Matrix.v[2].n[2] = (Fixed32) LUT16.e22;

       // Only operates if not identity...

       if (!MAT3isIdentity(&NewLUT -> Matrix, 0.0001)) {

              NewLUT -> wFlags |= LUT_HASMATRIX;
       }


       // Copy input tables

       AllLinear = 0;
       for (i=0; i < NewLUT -> InputChan; i++) {

              PtrW = (LPWORD) malloc(sizeof(WORD) * NewLUT -> InputEntries);
              NewLUT -> L1[i] = PtrW;
              fread(PtrW, sizeof(WORD), NewLUT -> InputEntries, Icc -> stream);
              AdjustEndianessArray16(PtrW, NewLUT -> InputEntries);
              AllLinear += cmsIsLinear(NewLUT -> L1[i], NewLUT -> InputEntries);
              }

       // Linear input, so ignore full step

       if (AllLinear == NewLUT -> InputChan) {

              NewLUT -> wFlags &= ~LUT_HASTL1;
       }


       // Copy 3D CLUT

       nTabSize = (NewLUT -> OutputChan * uipow(NewLUT->cLutPoints,
                                                NewLUT->InputChan));
       PtrW = (LPWORD) malloc(sizeof(WORD) * nTabSize);

       NewLUT -> T = PtrW;
       fread(PtrW, sizeof(WORD), nTabSize, Icc -> stream);
       AdjustEndianessArray16(NewLUT -> T, nTabSize);

       // Copy output tables

       AllLinear = 0;
       for (i=0; i < NewLUT -> OutputChan; i++) {

              PtrW = malloc(sizeof(WORD) * NewLUT -> OutputEntries);
              NewLUT -> L2[i] = PtrW;
              fread(PtrW, sizeof(WORD), NewLUT -> OutputEntries, Icc -> stream);
              AdjustEndianessArray16(PtrW, NewLUT -> OutputEntries);
              AllLinear += cmsIsLinear(NewLUT -> L2[i], NewLUT -> OutputEntries);
              }

       // Linear output, ignore step

       if (AllLinear == NewLUT -> InputChan)
       {
              NewLUT -> wFlags &= ~LUT_HASTL2;
       }


       cmsCalcL16Params(NewLUT -> InputEntries,  &NewLUT -> In16params);
       cmsCalcL16Params(NewLUT -> OutputEntries, &NewLUT -> Out16params);
       cmsCalcCLUT16Params(NewLUT -> cLutPoints,  NewLUT -> InputChan,
                                                  NewLUT -> OutputChan,
                                                  &NewLUT -> CLut16params);
}



LPLUT cmsReadICCLut(cmsHPROFILE hProfile, icTagSignature sig)
{

    LPICCPROFILE        Icc = (LPICCPROFILE) (LPSTR) hProfile;
    icTagBase           Base;
    int                 n;
    size_t              offset;
    LPLUT               NewLUT;

    n = SearchTag(Icc, sig);
    if (n < 0)
    {
    cmsSignalError(LCMS_ERRC_ABORTED, "Tag not found");
    return NULL;
    }

    // LUT only in files

    if (!Icc -> stream)
    {
    cmsSignalError(LCMS_ERRC_ABORTED, "Internal error");
    return NULL;
    }

    offset = Icc -> TagOffsets[n];

       
    if (fseek(Icc -> stream, offset, SEEK_SET))
    {
       cmsSignalError(LCMS_ERRC_ABORTED, "seek error; probably corrupted file");
       return NULL;
    }

    fread(&Base, sizeof(icTagBase), 1, Icc -> stream);
    AdjustEndianess32((LPBYTE) &Base.sig);

    NewLUT = cmsAllocLUT();
    if (!NewLUT)
    {
       cmsSignalError(LCMS_ERRC_ABORTED, "cmsAllocLUT() failed");
       return NULL;
    }


    switch (Base.sig) {

    case icSigLut8Type: ReadLUT8(Icc, NewLUT); break;
    case icSigLut16Type: ReadLUT16(Icc, NewLUT); break;

    default:  cmsSignalError(LCMS_ERRC_ABORTED, "Bad tag signature %lx found.", Base.sig);
              cmsFreeLUT(NewLUT);
              return NULL;
    }

    return NewLUT;
}


// Take an ASCII item

static
int ReadICCAscii(cmsHPROFILE hProfile, icTagSignature sig, char *Name)
{
    LPICCPROFILE        Icc = (LPICCPROFILE) (LPSTR) hProfile;
    icTagBase           Base;
    size_t              offset, size;
    int                 n;

    n = SearchTag(Icc, sig);
    if (n < 0)
    {
    cmsSignalError(LCMS_ERRC_ABORTED, "Tag not found");
    return -1;
    }


    if (!Icc -> stream)
    {
    CopyMemory(Name, Icc -> TagPtrs[n],
                     Icc -> TagSizes[n]);

    return Icc -> TagSizes[n];
    }

    offset = Icc -> TagOffsets[n];
    size   = Icc -> TagSizes[n];

    fseek(Icc -> stream, offset, SEEK_SET);
    fread(&Base, sizeof(icTagBase), 1, Icc -> stream);
    AdjustEndianess32((LPBYTE) &Base.sig);

    switch (Base.sig)
    {
    case icSigTextDescriptionType: {

           icUInt32Number  AsciiCount;

           fread(&AsciiCount, sizeof(icUInt32Number), 1, Icc -> stream);
           AdjustEndianess32((LPBYTE) &AsciiCount);
           fread(Name, 1, AsciiCount, Icc -> stream);
           size = AsciiCount;
           }
           break;

    case icSigTextType:

           size -= sizeof(icTagBase);
           fread(Name, 1, size, Icc -> stream);
           break;

    default:
              cmsSignalError(LCMS_ERRC_ABORTED, "Bad tag signature %lx found.", Base.sig);
              return -1;
    }

    return size;
}



// Take an XYZ item

static
int ReadICCXYZ(cmsHPROFILE hProfile, icTagSignature sig, LPcmsCIEXYZ Value)
{
    LPICCPROFILE        Icc = (LPICCPROFILE) (LPSTR) hProfile;
    icTagBase           Base;
    size_t              offset;
    int                 n;
    icXYZNumber         XYZ;

    n = SearchTag(Icc, sig);
    if (n < 0)
    {
    // Tag not found
    return -1;
    }


    if (!Icc -> stream)
    {
    CopyMemory(Value, Icc -> TagPtrs[n],
                      Icc -> TagSizes[n]);

    return Icc -> TagSizes[n];
    }

    offset = Icc -> TagOffsets[n];

    fseek(Icc -> stream, offset, SEEK_SET);
    fread(&Base, 1, sizeof(icTagBase), Icc -> stream);
    AdjustEndianess32((LPBYTE) &Base.sig);

    switch (Base.sig)
    {

    // Some apple broken embedded profiles does not have correct type

    case 0x7c3b10cL:
    case icSigXYZType:

           fread(&XYZ, sizeof(icXYZNumber), 1, Icc -> stream);
           Value -> X = Convert15Fixed16(XYZ.X);
           Value -> Y = Convert15Fixed16(XYZ.Y);
           Value -> Z = Convert15Fixed16(XYZ.Z);
           break;

    default:
           cmsSignalError(LCMS_ERRC_ABORTED, "Bad tag signature %lx found.", Base.sig);
           return -1;
    }

    return 1;
}



// White point must be in XYZ to avoid extra calculation on
// absolute intents

BOOL LCMSEXPORT cmsTakeMediaWhitePoint(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile)
{
       if (ReadICCXYZ(hProfile, icSigMediaWhitePointTag, Dest) < 0)
       {
              Dest->X = 0.964294;   // Default to D50
              Dest->Y = 1.0;
              Dest->Z = 0.825104;
              return FALSE;
       }
       return TRUE;
}


BOOL LCMSEXPORT cmsTakeMediaBlackPoint(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile)
{
       if (ReadICCXYZ(hProfile, icSigMediaBlackPointTag, Dest) < 0)
       {
              Dest->X = 0.0;
              Dest->Y = 0.0;
              Dest->Z = 0.0;
              return FALSE;
       }
       return TRUE;
}

BOOL  LCMSEXPORT cmsTakeIluminant(LPcmsCIEXYZ Dest, cmsHPROFILE hProfile)
{
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       Dest -> X = Icc -> Illuminant.X;
       Dest -> Y = Icc -> Illuminant.Y;
       Dest -> Z = Icc -> Illuminant.Z;

       return TRUE;
}

int LCMSEXPORT cmsTakeRenderingIntent(cmsHPROFILE hProfile)
{
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       return Icc -> RenderingIntent;
}


// Primaries are to be in xyY notation

BOOL LCMSEXPORT cmsTakeColorants(LPcmsCIEXYZTRIPLE Dest, cmsHPROFILE hProfile)
{

       if (ReadICCXYZ(hProfile, icSigRedColorantTag, &Dest -> Red) < 0) return FALSE;
       if (ReadICCXYZ(hProfile, icSigGreenColorantTag, &Dest -> Green) < 0) return FALSE;

       if (ReadICCXYZ(hProfile, icSigBlueColorantTag, &Dest -> Blue) < 0) return FALSE;



       return TRUE;

}


BOOL cdecl cmsReadICCMatrixRGB2XYZ(LPMAT3 r, cmsHPROFILE hProfile)
{
       cmsCIEXYZTRIPLE Primaries;

       if (!cmsTakeColorants(&Primaries, hProfile)) return FALSE;

       VEC3init(&r -> v[0], Primaries.Red.X, Primaries.Green.X,  Primaries.Blue.X);
       VEC3init(&r -> v[1], Primaries.Red.Y, Primaries.Green.Y,  Primaries.Blue.Y);
       VEC3init(&r -> v[2], Primaries.Red.Z, Primaries.Green.Z,  Primaries.Blue.Z);

       return TRUE;

}


LPGAMMATABLE cdecl cmsReadICCGamma(cmsHPROFILE hProfile, icTagSignature sig)
{
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;
       LPGAMMATABLE NewGamma;
       icUInt32Number Count;
       icTagBase      Base;
       size_t         offset;
       int            n;


       n = SearchTag(Icc, sig);
       if (n < 0)
       {
       cmsSignalError(LCMS_ERRC_ABORTED, "Tag not found");
       return NULL;
       }


       if (!Icc -> stream)
       {
       return DupGamma(Icc -> TagPtrs[n]);
       }

       offset = Icc -> TagOffsets[n];

       fseek(Icc -> stream, offset, SEEK_SET);
       fread(&Base, 1, sizeof(icTagBase), Icc -> stream);
       AdjustEndianess32((LPBYTE) &Base.sig);


       switch (Base.sig) {


       case icSigCurveType:

           fread(&Count, sizeof(icUInt32Number), 1, Icc -> stream);
           AdjustEndianess32((LPBYTE) &Count);


           switch (Count) {

           case 0:   // Linear.

                     NewGamma = cmsAllocGamma(2);
                     if (!NewGamma) return NULL;
                     NewGamma -> GammaTable[0] = 0;
                     NewGamma -> GammaTable[1] = 0xFFFF;
                     return NewGamma;

           case 1:  {
                     WORD SingleGammaFixed;

                     fread(&SingleGammaFixed, sizeof(WORD), 1, Icc -> stream);
                     return cmsBuildGamma(256, Convert8Fixed8(SingleGammaFixed));
                     }

           default:  // Curve

                     NewGamma = cmsAllocGamma(Count);
                     if (!NewGamma) return NULL;

                     fread(NewGamma -> GammaTable, sizeof(WORD), Count, Icc -> stream);
                     AdjustEndianessArray16(NewGamma -> GammaTable, Count);

                     return NewGamma;
              }

       default:
              cmsSignalError(LCMS_ERRC_ABORTED, "Bad tag signature %lx found.", Base.sig);
              return NULL;
       }

       // It would never reach here

       return NULL;
}



icColorSpaceSignature LCMSEXPORT cmsGetPCS(cmsHPROFILE hProfile)
{
       // Simply typecast
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       return Icc -> PCS;
}

icColorSpaceSignature LCMSEXPORT cmsGetColorSpace(cmsHPROFILE hProfile)
{
       // Simply typecast
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       return Icc -> ColorSpace;
}

// Check existance

BOOL LCMSEXPORT cmsIsTag(cmsHPROFILE hProfile, icTagSignature sig)
{
       // Simply typecast
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       return SearchTag(Icc, sig) >= 0;

}

icProfileClassSignature LCMSEXPORT cmsGetDeviceClass(cmsHPROFILE hProfile)
{
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       return Icc -> DeviceClass;
}




// We compute name with model - manufacturer

const char*  LCMSEXPORT cmsTakeProductName(cmsHPROFILE hProfile)
{
       static char Name[2048];
       char Manufacturer[512], Model[512];

       Name[0] = '\0';
       Manufacturer[0] = Model[0] = '\0';

       if (cmsIsTag(hProfile, icSigDeviceMfgDescTag))
       {
       ReadICCAscii(hProfile, icSigDeviceMfgDescTag, Manufacturer);
       }

       if (cmsIsTag(hProfile, icSigDeviceModelDescTag))
       {
       ReadICCAscii(hProfile, icSigDeviceModelDescTag, Model);
       }

       if (!Manufacturer[0] && !Model[0])
       {
              if (cmsIsTag(hProfile, icSigProfileDescriptionTag))
              {
              ReadICCAscii(hProfile, icSigProfileDescriptionTag, Name);
              return Name;
              }
              else return "{no name}";
       }


       if (!Manufacturer[0] || strncmp(Model, Manufacturer, 8) == 0)
              strcpy(Name, Model);
       else
              sprintf(Name, "%s - %s", Model, Manufacturer);

       return Name;

}


// We compute desc with manufacturer - model

const char*  LCMSEXPORT cmsTakeProductDesc(cmsHPROFILE hProfile)
{
       static char Name[2048];

       if (cmsIsTag(hProfile, icSigProfileDescriptionTag))
              {
              ReadICCAscii(hProfile, icSigProfileDescriptionTag, Name);
              }
       else return cmsTakeProductName(hProfile);

       if (strncmp(Name, "Copyrig", 7) == 0)
              return cmsTakeProductName(hProfile);

       return Name;
}


static
BOOL InRange(LPcmsCIExyY a, LPcmsCIExyY b, double tolerance)
{
       double dist_x, dist_y;

       dist_x = fabs(a->x - b->x);
       dist_y = fabs(a->y - b->y);

       return (tolerance >= dist_x * dist_x + dist_y * dist_y);

}


static
void IdentifyWhitePoint(char *Buffer, LPcmsCIEXYZ WhitePt)
{
       int i;
       cmsCIExyY Val;
       struct _Iluminants {
              const char *Name;
              cmsCIExyY Val;
              } SomeIluminants[] = {
                                   {"CIE illuminant A", {0.4476, 0.4074, 1.0}},
                                   {"CIE illuminant C", {0.3101, 0.3162, 1.0}},
                                   {"D50", {0.3457, 0.3585, 1.0}},
                                   {"D55", {0.3324, 0.3474, 1.0}},
                                   {"D65", {0.3127, 0.3290, 1.0}},
                                   {"D75", {0.2990, 0.3149, 1.0}}
                                   };

              cmsXYZ2xyY(&Val, WhitePt);

              for (i=0; i < 6; i++)
              {

                            if (InRange(&Val, &SomeIluminants[i].Val, 0.00001))
                            {
                                strcat(Buffer, "WhitePoint : ");
                                strcat(Buffer, SomeIluminants[i].Name);
                                strcat(Buffer, "\r\n\r\n");
                                return;
                            }
              }

              {
              char cBuff[300];

              sprintf(cBuff, "White point (X:%02.2g, Y:%02.2g, Z:%02.2g)\r\n\r\n",
                                          WhitePt -> X, WhitePt -> Y, WhitePt -> Z);
              strcat(Buffer, cBuff);
              }
}



const char*  LCMSEXPORT cmsTakeProductInfo(cmsHPROFILE hProfile)
{
       // Simply typecast
       LPICCPROFILE  Icc = (LPICCPROFILE) (LPSTR) hProfile;

       static char Info[4096];

       Info[0] = '\0';

       if (cmsIsTag(hProfile, icSigProfileDescriptionTag))
       {
       char Desc[1024];

       ReadICCAscii(hProfile, icSigProfileDescriptionTag, Desc);
       strcat(Info, Desc);
       strcat(Info, "\r\n\r\n");
       }


       if (cmsIsTag(hProfile, icSigCopyrightTag))
       {
       char Copyright[2048];

       ReadICCAscii(hProfile, icSigCopyrightTag, Copyright);
       strcat(Info, Copyright);
       strcat(Info, "\r\n\r\n");
       }



// KODAK private tag... But very useful

#define K007         0x4B303037

       // MonCal

       if (cmsIsTag(hProfile, K007))
       {
       char MonCal[1024];

       ReadICCAscii(hProfile, K007, MonCal);
       strcat(Info, MonCal);
       strcat(Info, "\r\n\r\n");
       }
       else
       {
       cmsCIEXYZ WhitePt;

       cmsTakeMediaWhitePoint(&WhitePt, hProfile);
       IdentifyWhitePoint(Info, &WhitePt);
       }


       if (Icc -> stream) {
              // strcat(Info, "Physical file :");
              strcat(Info, Icc -> PhysicalFile);
       }
       return Info;
}



